import pytest

from fastapi_keycloak import FastAPIKeycloak


class BaseTestClass:
    @pytest.fixture
    def idp(self):
        return FastAPIKeycloak(
            server_url="http://dev-keycloak.mydomain.test",
            client_id="test-client",
            client_secret="M48lQhr3xQs71LWOa4gVRDA3VYlwOyTw",
            admin_client_id="admin-cli",
            admin_client_secret="FOG8Q9iWLiIN6W8sE2v5pwuShKFO21fc",
            realm="test",
            callback_uri="http://localhost:8081/callback",
        )



def pytest_sessionstart(session):
    # subprocess.call(['sh', './start_infra.sh'])
    # print("Waiting for Keycloak to start")
    # sleep(60)  # Wait for startup
    pass


def pytest_sessionfinish(session):
    # subprocess.call(['sh', './stop_infra.sh'])
    pass



"""Keycloak API Client for integrating authentication and authorization with FastAPI"""

__version__ = "1.0.10.0"

from fastapi_keycloak_extended.api import FastAPIKeycloak
from fastapi_keycloak.model import (
    HTTPMethod,
    KeycloakError,
    KeycloakIdentityProvider,
    KeycloakRole,
    UsernamePassword,
)
from fastapi_keycloak_extended.model import (
    KeycloakToken,
    KeycloakGroup,
    OIDCUser,
    KeycloakUser,
    KeycloakRefreshToken,
)

__all__ = [
    FastAPIKeycloak.__name__,
    OIDCUser.__name__,
    UsernamePassword.__name__,
    HTTPMethod.__name__,
    KeycloakError.__name__,
    KeycloakUser.__name__,
    KeycloakToken.__name__,
    KeycloakRefreshToken.__name__,
    KeycloakRole.__name__,
    KeycloakIdentityProvider.__name__,
    KeycloakGroup.__name__,
]



from time import sleep
from typing import List

import httpretty as httpretty
import pytest as pytest
from fastapi import FastAPI
from fastapi.security import OAuth2PasswordBearer
from jose import JWTError
from requests import ReadTimeout

from fastapi_keycloak import HTTPMethod
from fastapi_keycloak.model import KeycloakRole
from tests import BaseTestClass


class TestAPIIntegration(BaseTestClass):
    def test_properties(self, idp):
        assert idp.public_key
        assert idp.admin_token
        assert idp.open_id_configuration
        assert idp.logout_uri
        assert idp.login_uri
        assert idp.roles_uri
        assert idp.token_uri
        assert idp.authorization_uri
        assert idp.user_auth_scheme
        assert idp.providers_uri
        assert idp.realm_uri
        assert idp.users_uri

    def test_admin_token(self, idp):
        assert idp.admin_token
        with pytest.raises(JWTError):  # Not enough segments
            idp.admin_token = "some rubbish"

        with pytest.raises(JWTError):  # Invalid crypto padding
            idp.admin_token = """
            eyJhbGciOiJIUzI1NiIsInR5cCI6IkpXVCJ9.eyJzdWIiOiIxMjM0NTY3ODkwIiwibmFtZSI6IkpvaG4gRG9lIiwiaWF0IjoxNTE2MjM5MDIyfQ.SflKxwRJSMeKKF2QT4fwpMeJf36POk6yJV_adQssw5c
            """

    def test_add_swagger_config(self, idp):
        app = FastAPI()
        assert app.swagger_ui_init_oauth is None
        idp.add_swagger_config(app)
        assert app.swagger_ui_init_oauth == {
            "usePkceWithAuthorizationCodeGrant": True,
            "clientId": idp.client_id,
            "clientSecret": idp.client_secret,
        }

    def test_user_auth_scheme(self, idp):
        assert isinstance(idp.user_auth_scheme, OAuth2PasswordBearer)

    def test_open_id_configuration(self, idp):
        assert idp.open_id_configuration
        assert type(idp.open_id_configuration) == dict

    def test_proxy(self, idp):
        response = idp.proxy(relative_path="/realms/Test", method=HTTPMethod.GET)
        assert type(response.json()) == dict

    @httpretty.activate(allow_net_connect=False)
    def test_timeout(self, idp):
        def request_callback(request, url, headers):
            sleep(1)
            return 200, headers, 'OK'

        httpretty.register_uri(httpretty.GET, f"{idp.server_url}/timeout", body=request_callback)
        idp.timeout = 0.5

        with pytest.raises(ReadTimeout):
            idp.proxy(relative_path="/timeout", method=HTTPMethod.GET)

    def test_get_all_roles_and_get_roles(self, idp):
        roles: List[KeycloakRole] = idp.get_all_roles()
        assert roles
        lookup = idp.get_roles(role_names=[role.name for role in roles])
        assert lookup
        assert len(roles) == len(lookup)

    def test_get_identity_providers(self, idp):
        assert idp.get_identity_providers() == []



class InvalidArgumentError(Exception):
    """Thrown error for invalid arguments.

    Attributes:
        status_code (int): The status code of the response received
        reason (str): The reason why the requests did fail
    """

    def __init__(self, status_code: int, reason: str):
        self.status_code = status_code
        self.reason = reason
        super().__init__(f"HTTP {status_code}: {reason}")



from typing import Any, Union

import requests
from fastapi import status
from fastapi.security import OAuth2PasswordBearer
from fastapi_keycloak import api
from fastapi_keycloak.exceptions import UserNotFound

from fastapi_keycloak_extended.exceptions import InvalidArgumentError
from fastapi_keycloak_extended.model import (
    KeycloakGroup,
    KeycloakRefreshToken,
    KeycloakToken,
    OIDCUser,
    KeycloakUser,
)


class FastAPIKeycloak(api.FastAPIKeycloak):
    """Instance to wrap the Keycloak API with FastAPI

    Attributes: _admin_token (KeycloakToken): A KeycloakToken instance, containing the access token that is used for
    any admin related request

    Example:
        ```python
        app = FastAPI()
        idp = KeycloakFastAPI(
            server_url="https://auth.some-domain.com/auth",
            client_id="some-test-client",
            client_secret="some-secret",
            admin_client_secret="some-admin-cli-secret",
            realm="Test",
            callback_uri=f"http://localhost:8081/callback"
        )
        idp.add_swagger_config(app)
        ```
    """

    def get_current_user(self, required_roles: list[str] = None) -> OIDCUser:
        """Returns the current user based on an access token in the HTTP-header. Optionally verifies roles are possessed
        by the user

        Args:
            required_roles List[str]: List of role names required for this endpoint

        Returns:
            OIDCUser: Decoded JWT content

        Raises:
            ExpiredSignatureError: If the token is expired (exp > datetime.now())
            JWTError: If decoding fails or the signature is invalid
            JWTClaimsError: If any claim is invalid
            HTTPException: If any role required is not contained within the roles of the users
        """

        def current_user(
            token: OAuth2PasswordBearer = api.Depends(self.user_auth_scheme),
        ) -> OIDCUser:
            """Decodes and verifies a JWT to get the current user

            Args:
                token OAuth2PasswordBearer: Access token in `Authorization` HTTP-header

            Returns:
                OIDCUser: Decoded JWT content

            Raises:
                ExpiredSignatureError: If the token is expired (exp > datetime.now())
                JWTError: If decoding fails or the signature is invalid
                JWTClaimsError: If any claim is invalid
                HTTPException: If any role required is not contained within the roles of the users
            """
            decoded_token = self._decode_token(token=token, audience="account")
            user = OIDCUser.parse_obj(decoded_token)
            if required_roles:
                for role in required_roles:
                    if role not in user.roles:
                        raise api.HTTPException(
                            status_code=403,
                            detail=f'Role "{role}" is required to perform this action',
                        )
            return user

        return current_user

    @api.result_or_error(response_model=KeycloakToken)
    def user_login(self, username: str, password: str) -> KeycloakToken:
        """Models the password OAuth2 flow. Exchanges username and password for an access token. Will raise detailed
        errors if login fails due to requiredActions

        Args:
            username (str): Username used for login
            password (str): Password of the user

        Returns:
            KeycloakToken: If the exchange succeeds

        Raises:
            HTTPException: If the credentials did not match any user
            MandatoryActionException: If the login is not possible due to mandatory actions
            KeycloakError: If the resulting response is not a successful HTTP-Code (>299, != 400, != 401)
            UpdateUserLocaleException: If the credentials we're correct but the has requiredActions of which the first one is to update his locale
            ConfigureTOTPException: If the credentials we're correct but the has requiredActions of which the first one is to configure TOTP
            VerifyEmailException: If the credentials we're correct but the has requiredActions of which the first one is to verify his email
            UpdatePasswordException: If the credentials we're correct but the has requiredActions of which the first one is to update his password
            UpdateProfileException: If the credentials we're correct but the has requiredActions of which the first one is to update his profile

        Notes:
            - To avoid calling this multiple times, you may want to check all requiredActions of the user if it fails due to a (sub)instance of an MandatoryActionException
        """
        headers = {"Content-Type": "application/x-www-form-urlencoded"}
        data = {
            "client_id": self.client_id,
            "client_secret": self.client_secret,
            "username": username,
            "password": password,
            "grant_type": "password",
        }
        response = requests.post(url=self.token_uri, headers=headers, data=data)
        if response.status_code == 401:
            raise api.HTTPException(status_code=401, detail="Invalid user credentials")
        if response.status_code == 400:
            user: KeycloakUser = self.get_user(query=f"username={username}")
            if len(user.requiredActions) > 0:
                reason = user.requiredActions[0]
                exception = {
                    "update_user_locale": api.UpdateUserLocaleException(),
                    "CONFIGURE_TOTP": api.ConfigureTOTPException(),
                    "VERIFY_EMAIL": api.VerifyEmailException(),
                    "UPDATE_PASSWORD": api.UpdatePasswordException(),
                    "UPDATE_PROFILE": api.UpdateProfileException(),
                }.get(
                    reason,  # Try to return the matching exception
                    # On custom or unknown actions return a MandatoryActionException by default
                    api.MandatoryActionException(
                        detail=f"This user can't login until the following action has been "
                        f"resolved: {reason}"
                    ),
                )
                raise exception
        return response

    @api.result_or_error(response_model=KeycloakGroup, is_list=True)
    def get_all_groups(self) -> list[KeycloakGroup]:
        """Get all base groups of the Keycloak realm

        Returns:
            List[KeycloakGroup]: All base groups of the realm

        Raises:
            KeycloakError: If the resulting response is not a successful HTTP-Code (>299)
        """
        return self._admin_request(
            url=f"{self.groups_uri}?briefRepresentation=false",
            method=api.HTTPMethod.GET,
        )

    @api.result_or_error(response_model=KeycloakGroup)
    def get_group_by_attribute(
        self, attribute: str, value: Any, search_in_subgroups=True
    ) -> list[KeycloakGroup]:
        """Return Group based on attribute

        Args:
            attribute (str): Attribute that should be looked up
            value (str): Value that should be looked up
            search_in_subgroups (bool): Whether to search in subgroups

        Returns:
            KeycloakGroup: Full entries stored at Keycloak. Or None if the path not found

        Raises:
            KeycloakError: If the resulting response is not a successful HTTP-Code (>299)
        """
        groups = self.get_all_groups()

        result = []
        for group in groups:
            if attribute in group.attributes and group.attributes[attribute] == value:
                result.append(group)
            elif search_in_subgroups and group.subGroups:
                for sub_group in group.subGroups:
                    if (
                        attribute in sub_group.attributes
                        and sub_group.attributes[attribute] == value
                    ):
                        result.append(sub_group)
        return result

    @api.result_or_error(response_model=KeycloakGroup)
    def get_group(self, group_id: str) -> KeycloakGroup or None:
        """Return Group based on group id

        Args:
            group_id (str): Group id to be found

        Returns:
             KeycloakGroup: Keycloak object by id. Or None if the id is invalid

        Notes:
            - The Keycloak RestAPI will only identify GroupRepresentations that
              use name AND id which is the only reason for existence of this function

        Raises:
            KeycloakError: If the resulting response is not a successful HTTP-Code (>299)
        """
        return self._admin_request(
            url=f"{self.groups_uri}/{group_id}?briefRepresentation=false",
            method=api.HTTPMethod.GET,
        )

    @api.result_or_error(response_model=KeycloakGroup)
    def create_group(
        self,
        group_name: str,
        parent: Union[KeycloakGroup, str] = None,
        attributes: dict = None,
    ) -> KeycloakGroup:
        """Create a group on the realm

        Args:
            group_name (str): Name of the new group
            parent (Union[KeycloakGroup, str]): Can contain an instance or object id
            attributes (dict): Custom datas of the new group

        Returns:
            KeycloakGroup: If creation succeeded, else it will return the error

        Raises:
            KeycloakError: If the resulting response is not a successful HTTP-Code (>299)
        """

        # If it's an object id get an instance of the object
        if isinstance(parent, str):
            parent = self.get_group(parent)

        if parent is not None:
            groups_uri = f"{self.groups_uri}/{parent.id}/children"
            path = f"{parent.path}/{group_name}"
        else:
            groups_uri = self.groups_uri
            path = f"/{group_name}"

        data = {"name": group_name}
        if attributes:
            data.update(attributes=attributes)

        response = self._admin_request(
            url=groups_uri, data=data, method=api.HTTPMethod.POST
        )
        if response.status_code == 201:
            return self.get_group_by_path(path=path, search_in_subgroups=True)
        else:
            return response

    @api.result_or_error(response_model=KeycloakUser)
    def create_user(
        self,
        first_name: str,
        last_name: str,
        username: str,
        email: str,
        password: str,
        enabled: bool = True,
        initial_roles: list[str] = None,
        send_email_verification: bool = True,
        attributes: dict = None,
    ) -> KeycloakUser:
        """

        Args:
            first_name (str): The first name of the new user
            last_name (str): The last name of the new user
            username (str): The username of the new user
            email (str): The email of the new user
            password (str): The password of the new user
            initial_roles (List[str]): The roles the user should posses. Defaults to `None`
            enabled (bool): True if the user should be able to be used. Defaults to `True`
            send_email_verification (bool): If true, the email verification will be added as an required
                                            action and the email triggered - if the user was created successfully.
                                            Defaults to `True`
            attributes (dict): The custom fields of new user

        Returns:
            KeycloakUser: If the creation succeeded

        Notes:
            - Also triggers the email verification email

        Raises:
            KeycloakError: If the resulting response is not a successful HTTP-Code (>299)
        """
        data = {
            "email": email,
            "username": username,
            "firstName": first_name,
            "lastName": last_name,
            "enabled": enabled,
            "credentials": [
                {"temporary": False, "type": "password", "value": password}
            ],
            "requiredActions": ["VERIFY_EMAIL" if send_email_verification else None],
        }
        if attributes:
            data.update(attributes=attributes)
        response = self._admin_request(
            url=self.users_uri, data=data, method=api.HTTPMethod.POST
        )
        if response.status_code != 201:
            return response
        user = self.get_user(query=f"username={username}")
        if send_email_verification:
            self.send_email_verification(user.id)
        if initial_roles:
            self.add_user_roles(initial_roles, user.id)
            user = self.get_user(user_id=user.id)
        return user

    @api.result_or_error(response_model=KeycloakUser)
    def get_user(self, user_id: str = None, query: str = "") -> KeycloakUser:
        """Queries the keycloak API for a specific user either based on its ID or any **native** attribute

        Args:
            user_id (str): The user ID of interest
            query: Query string. e.g. `email=testuser@codespecialist.com` or `username=codespecialist`

        Returns:
            KeycloakUser: If the user was found

        Raises:
            KeycloakError: If the resulting response is not a successful HTTP-Code (>299)
        """
        if user_id is None and (query is None or not query):
            raise InvalidArgumentError(
                status_code=400,
                reason=f"Make sure you have provided convenient arguments (query: {query} and user: {user_id}).",
            )
        if user_id is None:
            response = self._admin_request(
                url=f"{self.users_uri}?{query}&briefRepresentation=false",
                method=api.HTTPMethod.GET,
            )
            if not response.json():
                raise UserNotFound(
                    status_code=status.HTTP_404_NOT_FOUND,
                    reason=f"User query with filters of [{query}] did no match any users",
                )
            return KeycloakUser(**response.json()[0])
        else:
            response = self._admin_request(
                url=f"{self.users_uri}/{user_id}?briefRepresentation=false",
                method=api.HTTPMethod.GET,
            )
            if response.status_code == status.HTTP_404_NOT_FOUND:
                raise UserNotFound(
                    status_code=status.HTTP_404_NOT_FOUND,
                    reason=f"User with user_id[{user_id}] was not found",
                )
            return KeycloakUser(**response.json())

    @api.result_or_error(response_model=KeycloakUser)
    def update_user(self, user: KeycloakUser):
        """Updates a user. Requires the whole object.

        Args:
            user (KeycloakUser): The (new) user object

        Returns:
            KeycloakUser: The updated user

        Raises:
            KeycloakError: If the resulting response is not a successful HTTP-Code (>299)

        Notes: - You may alter any aspect of the user object, also the requiredActions for instance. There is no
        explicit function for updating those as it is a user update in essence
        """
        response = self._admin_request(
            url=f"{self.users_uri}/{user.id}",
            data=user.__dict__,
            method=api.HTTPMethod.PUT,
        )
        if response.status_code == 204:  # Update successful
            return self.get_user(user_id=user.id)
        return response

    @api.result_or_error(response_model=KeycloakUser, is_list=True)
    def get_all_users(self) -> list[KeycloakUser]:
        """Returns all users of the realm

        Returns:
            List[KeycloakUser]: All Keycloak users of the realm

        Raises:
            KeycloakError: If the resulting response is not a successful HTTP-Code (>299)
        """
        return self._admin_request(
            url=f"{self.users_uri}?briefRepresentation=false", method=api.HTTPMethod.GET
        )

    @api.result_or_error(response_model=KeycloakRefreshToken)
    def refresh_token(self, refresh_token: str) -> KeycloakRefreshToken:
        """
        Try to retrieve a new valid token from refresh token

        Args:
            refresh_token (str): refresh token

        Returns:
            KeycloakToken: If the exchange succeeds
        """
        headers = {"Content-Type": "application/x-www-form-urlencoded"}
        data = {
            "client_id": self.client_id,
            "client_secret": self.client_secret,
            "refresh_token": refresh_token,
            "grant_type": "refresh_token",
        }
        response = requests.post(url=self.token_uri, headers=headers, data=data)
        return response



from typing import List, Optional

import uvicorn
from fastapi import FastAPI, Depends, Query, Body, Request
from fastapi.responses import JSONResponse
from pydantic import SecretStr

from fastapi_keycloak_extended import (
    FastAPIKeycloak,
    HTTPMethod,
    KeycloakUser,
    OIDCUser,
    UsernamePassword,
    KeycloakError
)

app = FastAPI()
idp = FastAPIKeycloak(
    server_url="http://dev-keycloak.mydomain.test",
    client_id="test-client",
    client_secret="M48lQhr3xQs71LWOa4gVRDA3VYlwOyTw",
    admin_client_id="admin-cli",
    admin_client_secret="FOG8Q9iWLiIN6W8sE2v5pwuShKFO21fc",
    realm="test",
    callback_uri="http://localhost:8081/callback",
)
idp.add_swagger_config(app)


# Custom error handler for showing Keycloak errors on FastAPI
@app.exception_handler(KeycloakError)
async def keycloak_exception_handler(request: Request, exc: KeycloakError):
    return JSONResponse(
        status_code=exc.status_code,
        content={"message": exc.reason},
    )

    
# Admin


@app.post("/proxy", tags=["admin-cli"])
def proxy_admin_request(
    relative_path: str,
    method: HTTPMethod,
    additional_headers: dict = Body(None),
    payload: dict = Body(None),
):
    return idp.proxy(
        additional_headers=additional_headers,
        relative_path=relative_path,
        method=method,
        payload=payload,
    )


@app.get("/identity-providers", tags=["admin-cli"])
def get_identity_providers():
    return idp.get_identity_providers()


@app.get("/idp-configuration", tags=["admin-cli"])
def get_idp_config():
    return idp.open_id_configuration


# User Management


@app.get("/users", tags=["user-management"])
def get_users():
    return idp.get_all_users()


@app.get("/user", tags=["user-management"])
def get_user_by_query(query: str = None):
    return idp.get_user(query=query)


@app.post("/users", tags=["user-management"])
def create_user(
    first_name: str, last_name: str, email: str, password: SecretStr, id: str = None
):
    return idp.create_user(
        first_name=first_name,
        last_name=last_name,
        username=email,
        email=email,
        password=password.get_secret_value(),
        id=id,
    )


@app.get("/user/{user_id}", tags=["user-management"])
def get_user(user_id: str = None):
    return idp.get_user(user_id=user_id)


@app.put("/user", tags=["user-management"])
def update_user(user: KeycloakUser):
    return idp.update_user(user=user)


@app.delete("/user/{user_id}", tags=["user-management"])
def delete_user(user_id: str):
    return idp.delete_user(user_id=user_id)


@app.put("/user/{user_id}/change-password", tags=["user-management"])
def change_password(user_id: str, new_password: SecretStr):
    return idp.change_password(user_id=user_id, new_password=new_password)


@app.put("/user/{user_id}/send-email-verification", tags=["user-management"])
def send_email_verification(user_id: str):
    return idp.send_email_verification(user_id=user_id)


# Role Management


@app.get("/roles", tags=["role-management"])
def get_all_roles():
    return idp.get_all_roles()


@app.get("/role/{role_name}", tags=["role-management"])
def get_role(role_name: str):
    return idp.get_roles([role_name])


@app.post("/roles", tags=["role-management"])
def add_role(role_name: str):
    return idp.create_role(role_name=role_name)


@app.delete("/roles", tags=["role-management"])
def delete_roles(role_name: str):
    return idp.delete_role(role_name=role_name)


# Group Management


@app.get("/groups", tags=["group-management"])
def get_all_groups():
    return idp.get_all_groups()


@app.get("/group/{group_name}", tags=["group-management"])
def get_group(group_name: str):
    return idp.get_groups([group_name])


@app.get("/group-by-path/{path: path}", tags=["group-management"])
def get_group_by_path(path: str):
    return idp.get_group_by_path(path)


@app.post("/groups", tags=["group-management"])
def add_group(group_name: str, parent_id: Optional[str] = None):
    return idp.create_group(group_name=group_name, parent=parent_id)


@app.delete("/groups", tags=["group-management"])
def delete_groups(group_id: str):
    return idp.delete_group(group_id=group_id)


# User Roles


@app.post("/users/{user_id}/roles", tags=["user-roles"])
def add_roles_to_user(user_id: str, roles: Optional[List[str]] = Query(None)):
    return idp.add_user_roles(user_id=user_id, roles=roles)


@app.get("/users/{user_id}/roles", tags=["user-roles"])
def get_user_roles(user_id: str):
    return idp.get_user_roles(user_id=user_id)


@app.delete("/users/{user_id}/roles", tags=["user-roles"])
def delete_roles_from_user(user_id: str, roles: Optional[List[str]] = Query(None)):
    return idp.remove_user_roles(user_id=user_id, roles=roles)


# User Groups


@app.post("/users/{user_id}/groups", tags=["user-groups"])
def add_group_to_user(user_id: str, group_id: str):
    return idp.add_user_group(user_id=user_id, group_id=group_id)


@app.get("/users/{user_id}/groups", tags=["user-groups"])
def get_user_groups(user_id: str):
    return idp.get_user_groups(user_id=user_id)


@app.delete("/users/{user_id}/groups", tags=["user-groups"])
def delete_groups_from_user(user_id: str, group_id: str):
    return idp.remove_user_group(user_id=user_id, group_id=group_id)


# Example User Requests


@app.get("/protected", tags=["example-user-request"])
def protected(user: OIDCUser = Depends(idp.get_current_user())):
    return user


@app.get("/current_user/roles", tags=["example-user-request"])
def get_current_users_roles(user: OIDCUser = Depends(idp.get_current_user())):
    return user.roles


@app.get("/admin", tags=["example-user-request"])
def company_admin(
    user: OIDCUser = Depends(idp.get_current_user(required_roles=["admin"])),
):
    return f"Hi admin {user}"


@app.post("/login", tags=["example-user-request"])
def login(user: UsernamePassword = Body(...)):
    return idp.user_login(
        username=user.username, password=user.password.get_secret_value()
    )


# Auth Flow


@app.get("/login-link", tags=["auth-flow"])
def login_redirect():
    return idp.login_uri


@app.get("/callback", tags=["auth-flow"])
def callback(session_state: str, code: str):
    return idp.exchange_authorization_code(session_state=session_state, code=code)


@app.get("/logout", tags=["auth-flow"])
def logout():
    return idp.logout_uri


if __name__ == "__main__":
    uvicorn.run("app:app", host="127.0.0.1", port=8081)
